package net.sys42.dakedres.grafting.block.entity;

import net.minecraft.block.BlockState;
import net.minecraft.block.entity.BlockEntity;
import net.minecraft.block.entity.BlockEntityType;
import net.minecraft.inventory.Inventory;
import net.minecraft.item.ItemStack;
import net.minecraft.nbt.NbtCompound;
import net.minecraft.nbt.NbtList;
import net.minecraft.registry.RegistryWrapper;
import net.minecraft.util.math.BlockPos;

import java.util.ArrayList;
import java.util.List;

public abstract class OrderedContainer extends BlockEntity implements Inventory {
	private final int size;
	private final ArrayList<ItemStack> inventory;
	public OrderedContainer(BlockEntityType<?> type, BlockPos pos, BlockState state, int size) {
		super(type, pos, state);
		this.size = size;
		this.inventory = new ArrayList<>(size);
	}

	protected void readNbt(NbtCompound nbt, RegistryWrapper.WrapperLookup registries) {
		super.readNbt(nbt, registries);
		readInventoryNbt(nbt, registries);
	}

	private void readInventoryNbt(NbtCompound nbt, RegistryWrapper.WrapperLookup registries) {
		NbtList nbtList = nbt.getList("Items", 10);

		for(int i = 0; i < nbtList.size(); ++i) {
			NbtCompound nbtCompound = nbtList.getCompound(i);
			inventory.add((ItemStack) ItemStack.fromNbt(registries, nbtCompound).orElse(ItemStack.EMPTY));
		}
	}

	protected void writeNbt(NbtCompound nbt, RegistryWrapper.WrapperLookup registries) {
		super.writeNbt(nbt, registries);
		writeInventoryNbt(nbt, registries);
	}

	private void writeInventoryNbt(NbtCompound nbt, RegistryWrapper.WrapperLookup registries) {
		NbtList nbtList = new NbtList();

		for (ItemStack itemStack : inventory) {
			if (!itemStack.isEmpty()) {
				NbtCompound nbtCompound = new NbtCompound();
				nbtList.add(itemStack.encode(registries, nbtCompound));
			}
		}

		nbt.put("Items", nbtList);
	}

	@Override
	public int size() {
		return size;
	}

	@Override
	public boolean isEmpty() {
		return inventory.isEmpty();
	}

	public int filledSlots() {
		return inventory.size();
	}

	public boolean hasRoom() {
		return filledSlots() < size;
	}

	private boolean validSlot(int slot) {
		return slot >= 0 && slot < filledSlots();
	}

	@Override
	public ItemStack getStack(int slot) {
		if(validSlot(slot)) {
			return inventory.get(slot);
		} else {
			return ItemStack.EMPTY;
		}
	}

	// Sometimes Inventory#removeStack will be called in response to the
	// count being 0, but not always. Here we check if that is the case
	// before pulling the newest item.
	//
	// This is undesired behavior (as it allows the user to remove items
	// that are not at the top of the stack), and should be avoided. This
	// method prevents the side effects from being even worse, however.
	private ItemStack checkSlot(int slot) {
		if(!validSlot(slot)) {
			return ItemStack.EMPTY;
		}

		ItemStack stack = this.getStack(slot);

		if(stack.isEmpty()) {
			inventory.remove(slot);
			this.markDirty();
			return stack;
		}

		return stack;
	}

	@Override
	public ItemStack removeStack(int slot, int amount) {
		ItemStack stack = checkSlot(slot);

		if(stack.isEmpty() || amount <= 0) {
			return stack;
		}

		if(amount <= stack.getCount()) {
			return stack.split(amount);
		}

		return  ItemStack.EMPTY;
	}

	@Override
	public ItemStack removeStack(int slot) {
		ItemStack stack = checkSlot(slot);

		if(stack.isEmpty()) {
			return stack;
		}

		return pullStack();
	}

	@Override
	public void setStack(int slot, ItemStack stack) {
		if(stack.isEmpty()) {
			if(validSlot(slot) && this.getStack(slot).isEmpty()) {
				inventory.remove(slot);
				this.markDirty();
			}
			return;
		}

		if(slot >= this.filledSlots()) {
			pushStack(stack);
		} else {
			this.inventory.set(slot, stack);
		}

		markDirty();
	}

	public ItemStack pushStack(ItemStack stack) {
		if(stack.isEmpty()) {
			return stack;
		}

		if(hasRoom()) {
			inventory.add(stack);
			distributeStack(stack);
			markDirty();
			return ItemStack.EMPTY;
		}

		return stack;
	}

	public ItemStack pullStack() {
		if(this.filledSlots() > 0) {
			ItemStack o = inventory.removeLast();
			markDirty();
			return o;
		} else {
			return ItemStack.EMPTY;
		}
	}

	public boolean hasItemsLike(ItemStack stack) {
		for(ItemStack itemStack : inventory) {
			if(ItemStack.areItemsAndComponentsEqual(itemStack, stack)) {
				return true;
			}
		}

		return false;
	}

	@Override
	public void clear() {
		inventory.clear();
	}

	public List<ItemStack> getHeldStacks() {
		return inventory;
	}

	private void clean() {
		for(int i = 0; i < inventory.size(); i++) {
			if(inventory.get(i).isEmpty()) {
				inventory.remove(i);
				i--;
			}
		}
	}

	//TODO: Implement according to LockedContainer
	@Override
	public void markDirty() {
		clean();
		super.markDirty();
	}

	public ItemStack topStack() {
		return this.inventory.getLast();
	}

//	public ItemStack topStackAlike(ItemStack comparisonStack) {
//		for(ItemStack stack : inventory) {
//			if(ItemStack.areItemsAndComponentsEqual(stack, comparisonStack)) {
//				return stack;
//			}
//		}
//
//		return ItemStack.EMPTY;
//	}

	public List<ItemStack> similarStacks(ItemStack comparisonStack) {
		var out = new ArrayList<ItemStack>();

		for(int i = 0; i < filledSlots(); i++) {
			ItemStack stack = getStack(i);

			if(ItemStack.areItemsAndComponentsEqual(stack, comparisonStack)) {
				out.add(stack);
			}
		}

		return out;
	}

	public void distributeStack(ItemStack target) {
		distributeStack(target, 0);
	}

	public void distributeStack(ItemStack target, int amount) {
		List<ItemStack> candidates = similarStacks(target);

		if(candidates.isEmpty()) {
			return;
		}

		int total = amount;
		int maxTotal = 0;

		for(ItemStack candidate : candidates) {
			total += candidate.getCount();
			maxTotal += candidate.getMaxCount();
			candidate.setCount(0);
		}

		int remainder = 0;
		if(total > maxTotal) {
			remainder = total - maxTotal;
			total = maxTotal;
		}

		int share = total / candidates.size();

		for(ItemStack candidate : candidates) {
			candidate.setCount(share);
			total -= share;
		}

		int i = 0;
		while(total > 0) {
			candidates.get(i % candidates.size()).increment(1);
			i++;
			total--;
		}

		if(amount != 0) {
			target.decrement(amount);
		}
	}
}
